import bpy
from mathutils import Matrix, Vector
from bpy.types import Operator, GizmoGroup
from ..utils.poll import poll_placement
from ..matrix.matrix_ensure import get_orient_from_type
from .. import var
from bpy.props import FloatVectorProperty, FloatProperty, IntProperty
from ..utils.draw_utils.through import alpha_gizmo, alpha_arc_gizmo

from mathutils.geometry import intersect_point_line
from bpy_extras.view3d_utils import (
    location_3d_to_region_2d,
    region_2d_to_location_3d,
    region_2d_to_vector_3d,
    region_2d_to_origin_3d,
)
from gpu import state
from ..utils.draw_utils.draw import draw_batch, draw_custom_batch_mw
import gpu
from gpu_extras.batch import batch_for_shader
from numpy import sign
from ..keymaps import keymaps_gizmo_tweak
from .. import get_prefs



class GP_OT_placement_move(Operator):
    bl_idname = 'gp.placement_move'
    bl_label = 'Placement'
    bl_description = ('Gizmo to place the selected by normals\n'
                      '• Shift - Duplicate Move')

    orient_matrix: FloatVectorProperty(size=(3, 3), subtype='MATRIX')
    center_override: FloatVectorProperty(size=3, default=[0, 0, 0])
    duplicate = False
    use_standard_options = False

    def execute(self, context):
        # Сохраняем значения опций Snap и Align Rotation
        use_snap = context.scene.tool_settings.use_snap
        use_snap_align_rotation = context.scene.tool_settings.use_snap_align_rotation
        snap_elements = context.scene.tool_settings.snap_elements
        snap_target = context.scene.tool_settings.snap_target

        # Устанавливаем нужные значения опций Snap и Align Rotation
        context.scene.tool_settings.use_snap = not self.use_standard_options
        context.scene.tool_settings.use_snap_align_rotation = True

        if self.use_standard_options is False: # 'FACE' not in context.scene.tool_settings.snap_elements
            context.scene.tool_settings.snap_elements = {'FACE'}
        context.scene.tool_settings.snap_target = 'CENTER'

        # Вызываем оператор bpy.ops.transform.transform
        if self.duplicate:
            bpy.ops.object.duplicate_move(
                    'INVOKE_DEFAULT',
                    TRANSFORM_OT_translate = {
                        'orient_matrix': self.orient_matrix,
                        'release_confirm': True,
                        },
                    )
        else:
            bpy.ops.transform.transform(
                    'INVOKE_DEFAULT',
                    mode = 'TRANSLATION',
                    release_confirm = True,
                    orient_matrix = self.orient_matrix,
                    center_override = self.center_override,
                    )

        # Возвращаем опции Snap и Align Rotation к сохраненным значениям
        if self.use_standard_options:
            context.scene.tool_settings.use_snap = not use_snap
        else:
            context.scene.tool_settings.use_snap = use_snap
        context.scene.tool_settings.use_snap_align_rotation = use_snap_align_rotation
        context.scene.tool_settings.snap_elements = snap_elements
        context.scene.tool_settings.snap_target = snap_target
        
        return {'FINISHED'}

    def invoke(self, context, event):
        self.duplicate = event.shift
        self.use_standard_options = event.ctrl
        return self.execute(context)



class GP_OT_placement_move_z(Operator):
    bl_idname = 'gp.placement_move_z'
    bl_label = 'Placement'
    bl_description = ('Moving in the local Z axis\n'
                      '• Shift - Moving object data')
    bl_options = {'REGISTER', 'UNDO'}

    orient_matrix: FloatVectorProperty(size=(3, 3), subtype='MATRIX')
    center_override: FloatVectorProperty(size=3, default=[0, 0, 0])
    

    _closest_point = Vector()
    _initial_position = Vector()
    _verts = {}
    _obj = None
   

    def modal(self, context, event):
        if event.type == 'MOUSEMOVE':
            # --- Получение точки пересечения
            mouse_pos = Vector((event.mouse_region_x, event.mouse_region_y))
            mouse_pos_3d = region_2d_to_location_3d(context.region, context.region_data, mouse_pos, self.object_location)
            end_line = self.object_location + self.object_orientation @ Vector((0.0, 0.0, 1.0)) # Создаем линию по оси Z объекта
            self._closest_point = intersect_point_line(mouse_pos_3d, self.object_location, end_line)[0]
            
            # --- Расчет смещения
            direction_vector = self.object_location - self._initial_position
            direction_vector.normalize()
            current_vector = self._closest_point - self._initial_position
            dot_product = direction_vector.dot(current_vector)

            value = dot_product
            if event.shift:
                value = dot_product
            else:
                value = dot_product * 0.1
            if event.ctrl:
                if event.shift:
                    value = round(dot_product, 1)
                else:
                    value = round(dot_product)
            
            # --- Применяем изменение к смещению по оси Z
            mw = self._obj.matrix_world.inverted()
            for v in self._obj.data.vertices:
                v.co.z = (mw @ self._verts[v.index])[2] + value

            # --- Обновление сцены, чтобы увидеть изменения
            #context.view_layer.update()
        
        elif event.type in {'LEFTMOUSE', 'RET', 'NUMPAD_ENTER'}:  # Завершение оператора
            #bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            mw = self._obj.matrix_world.inverted()
            for v in self._obj.data.vertices:
                v.co.z = (mw @ self._verts[v.index])[2]
            #bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'CANCELLED'}

        return{'PASS_THROUGH'}


    def invoke(self, context, event):
        if event.shift and context.mode == 'OBJECT':
            """ self._handle = bpy.types.SpaceView3D.draw_handler_add(
                self.draw_callback_px, (context,),
                'WINDOW', 'POST_VIEW'
            ) """

            self.object_orientation = context.object.matrix_world.to_3x3()
            self.object_location = context.object.matrix_world.to_translation()

            mouse_pos = Vector((event.mouse_region_x, event.mouse_region_y))
            mouse_pos_3d = region_2d_to_location_3d(context.region, context.region_data, mouse_pos, self.object_location)
            start_line = self.object_location 
            end_line = start_line + self.object_orientation @ Vector((0.0, 0.0, 1.0))
            self._initial_position = intersect_point_line(mouse_pos_3d, start_line, end_line)[0]

            self._obj = context.active_object
            mw = self._obj.matrix_world
            for v in self._obj.data.vertices:
                self._verts[v.index] = mw @ v.co

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            bpy.ops.transform.transform(
                    'INVOKE_DEFAULT',
                    mode = 'TRANSLATION',
                    release_confirm = True,
                    constraint_axis = (False, False, True),
                    orient_matrix = self.orient_matrix,
                    center_override = self.center_override,
                    )
            return {'FINISHED'}

    """ def draw_callback_px(self, context):
        # 50% alpha, 2 pixel width line
        shader = gpu.shader.from_builtin('UNIFORM_COLOR')
        gpu.state.blend_set('ALPHA')
        gpu.state.line_width_set(5.0)
        gpu.state.point_size_set(10.0)

        batch = batch_for_shader(shader, 'POINTS', {"pos": [self._initial_position, self._closest_point]})
        shader.uniform_float("color", (0.0, 0.0, 1.0, 1.0))
        batch.draw(shader)

        # restore opengl defaults
        gpu.state.point_size_set(1.0)
        gpu.state.line_width_set(1.0)
        gpu.state.blend_set('NONE') """




class GP_GGT_placement(GizmoGroup):
    
    bl_idname = 'GP_GGT_placement'
    bl_label = 'Gizmo Placement'
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW' 
    bl_options = {'3D', 'PERSISTENT', 'SHOW_MODAL_ALL'} #'DEPTH_3D' , 'TOOL_INIT', 'SELECT', , 'SCALE'  , 'SHOW_MODAL_ALL' , 'PERSISTENT' , VR_REDRAWS 
    

    @classmethod
    def poll(cls, context):
        return poll_placement(context)
    
    
    @classmethod
    def setup_keymap(cls, keyconfig):
        return keymaps_gizmo_tweak(keyconfig)

    
    def setup(self, context):
        # --- MOVE
        self.arrow = self.gizmos.new('GIZMO_GT_arrow_3d')
        self.arrow.use_draw_offset_scale = True 
        self.arrow.use_tooltip = False
        self.arrow.transform = {'CONSTRAIN'}
        self.arrow_op = self.arrow.target_set_operator('gp.placement_move_z')
    
        # --- SCALE
        self.scale = self.gizmos.new('GIZMO_GT_arrow_3d')
        self.scale.draw_style = 'BOX'
        self.scale.use_draw_offset_scale = True
        self.scale.use_tooltip = False
        self.scale_op = self.scale.target_set_operator('transform.transform')
        self.scale_op.mode = 'RESIZE'
        self.scale_op.release_confirm = True
        self.scale_op.constraint_axis = (False, False, True)
        #self.scale_op.snap = False

        # --- ROTATE
        self.dial = self.gizmos.new('GIZMO_GT_dial_3d')
        self.dial.draw_options = {'CLIP'}
        self.dial.use_tooltip = False
        self.dial_op = self.dial.target_set_operator('transform.transform')
        self.dial_op.mode = 'ROTATION'
        self.dial_op.release_confirm = True
        self.dial_op.constraint_axis = (False, False, True)
        self.dial_op.snap = False

        # --- DOT MOVE
        self.dot = self.gizmos.new('GIZMO_GT_move_3d')
        self.dot.draw_options = {'FILL_SELECT'}
        #self.dot.use_tooltip = False
        self.dot_op = self.dot.target_set_operator('gp.placement_move')


    def invoke_prepare(self, context, gizmo):
        matrix = gizmo.matrix_basis.copy().to_3x3()
        pos = gizmo.matrix_basis.translation.copy()
        self.arrow_op.orient_matrix = matrix
        self.arrow_op.center_override = pos
        self.scale_op.orient_matrix = matrix
        self.scale_op.center_override = pos
        self.dial_op.orient_matrix = matrix
        self.dial_op.center_override = pos
        self.dot_op.orient_matrix = matrix
        self.dot_op.center_override = pos


    def refresh(self, context):
        mm, ms = get_orient_from_type(position='ACTIVE_ELEMENT', orient='LOCAL', type='PLACEMENT')
        self.dot.matrix_basis = mm
        self.arrow.matrix_basis = mm
        self.dial.matrix_basis = mm
        self.scale.matrix_basis = ms
        

    def draw_prepare(self, context):
        props = get_prefs()
        settings = context.scene.GP_scene_set
        
        coef = 1.5
        l_size_a = props.size_gizmo_placement * coef
        l_size_d = props.size_gizmo_placement * 0.8 * coef
        l_dot_size = props.size_gizmo_placement * 0.2 * coef
        lw = props.size_gizmo_placement * 5.0 #* props.luna_lw
         
        # --- принудительное обновление матрицы
        orient_slots = context.window.scene.transform_orientation_slots[0].type
        if var.REFRESH_MAT or orient_slots=='VIEW' or settings.orientation=='VIEW' or settings.position=='VIEW':
            self.refresh(context)
        var.REFRESH_MAT = False
        
        alpha, hide_select = alpha_gizmo(self.dot.matrix_basis, props)
        alpha_d, hide_select_d = alpha_arc_gizmo(self.dot.matrix_basis, props)

        # --- MOVE
        arrow = self.arrow
        arrow.length = 0.8
        arrow.line_width = lw
        arrow.alpha = alpha
        arrow.hide_select = hide_select
        arrow.alpha_highlight = props.highlight_alpha_gizmo
        arrow.scale_basis = l_size_a
        arrow.color = (1.0, 1.0, 0.0)
        arrow.color_highlight = props.select_color
        arrow.matrix_offset = Matrix.Translation( Vector( (0.0, 0.0, 0.2) ) )
    
            
        # --- ROTATE
        dial = self.dial
        dial.line_width = lw
        dial.alpha = alpha_d
        dial.hide_select = hide_select_d
        dial.alpha_highlight = props.highlight_alpha_gizmo
        dial.scale_basis = l_size_d
        dial.color = (0.0, 0.27, 1.0)
        dial.color_highlight = props.select_color
        

        # --- SCALE
        scale = self.scale
        scale.line_width = lw
        scale.length = 0.8
        scale.alpha = alpha
        scale.hide_select = hide_select
        scale.alpha_highlight = props.highlight_alpha_gizmo
        scale.scale_basis = l_size_a
        scale.color = (0.0, 0.27, 1.0)
        scale.color_highlight = props.select_color
        scale.matrix_offset = Matrix.Translation( Vector( (0.0, 0.0, 0.2) ) )


        # --- DOT
        dot = self.dot
        dot.color = props.elements_color
        dot.color_highlight = props.select_color
        dot.alpha = alpha_d
        dot.hide_select = hide_select_d
        dot.alpha_highlight = props.highlight_alpha_gizmo
        dot.line_width = lw
        dot.scale_basis = l_dot_size
      


classes = [
    GP_OT_placement_move,
    GP_GGT_placement,
    GP_OT_placement_move_z,
    ]


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)