import bpy
from mathutils import Matrix, Quaternion, Vector
from math import radians


def get_global_orient():
    x = Quaternion((0.0, 1.0, 0.0), radians(90)) 
    y = Quaternion((1.0, 0.0, 0.0), radians(-90))
    z = Quaternion((0.0, 0.0, 1.0), radians(-90))
    return x, y, z


def get_local_orient():
    rot = bpy.context.object.matrix_world.decompose()[1]
    x = rot @ Quaternion((0.0, 1.0, 0.0), radians(90))
    y = rot @ Quaternion((1.0, 0.0, 0.0), radians(-90))
    z = rot @ Quaternion((0.0, 0.0, 1.0), radians(-90))
    return x, y, z



def get_gimbal_orient():
    ob = bpy.context.object
    rot_mode = ob.rotation_mode

    if rot_mode == 'QUATERNION':
        x = Quaternion((0.0, 1.0, 0.0), radians(90)) 
        y = Quaternion((1.0, 0.0, 0.0), radians(-90))
        z = Quaternion((0.0, 0.0, 1.0), radians(-90))
    elif rot_mode in ['XYZ', 'XZY', 'YXZ', 'YZX', 'ZXY', 'ZYX']: # FIXME все euler modes (сейчас только XYZ)
        rot = ob.matrix_world.decompose()[1]
        x = rot @ Quaternion((0.0, 1.0, 0.0), radians(90)) 
        y = Quaternion((0.0, 0.0, 1.0), ob.rotation_euler[2]) @ Quaternion((1.0, 0.0, 0.0), radians(-90))
        z = Quaternion((0.0, 0.0, 1.0), radians(-90))
    elif rot_mode == 'AXIS_ANGLE':  # FIXME angle mode
        rot = ob.rotation_axis_angle
        axis_angle_rot = Quaternion(rot[1:], rot[0])
        x = axis_angle_rot @ Quaternion((0.0, 1.0, 0.0), radians(90))
        y = Quaternion((0.0, 0.0, 1.0), radians(-90)) @ axis_angle_rot
        z = Quaternion((0.0, 0.0, 1.0), radians(-90))

    return x, y, z

"""  xD = rot @ Quaternion( (0.0, 1.0, 0.0), radians(-90) ) @ Quaternion( (0.0, 0.0, 1.0), radians(props.arc_angle[0]) )
yD = Quaternion( (0.0, 0.0, 1.0), ob.rotation_euler[2] ) @ Quaternion( (1.0, 0.0, 0.0), radians(90) ) @ Quaternion( (0.0, 0.0, 1.0), radians(props.arc_angle[1]) )
zD = Quaternion( (0.0, 0.0, 1.0), radians(props.arc_angle[2]) ) """


def get_view_orient():
    view_inv = bpy.context.region_data.view_matrix.inverted()
    rot = view_inv.decompose()[1]
    x = rot @ Quaternion((0.0, 1.0, 0.0), radians(90)) 
    y = rot @ Quaternion((1.0, 0.0, 0.0), radians(-90))
    z = rot @ Quaternion((0.0, 0.0, 1.0), radians(-90))
    return x, y, z


def get_cursor_orient():
    cursor_mat = bpy.context.scene.cursor.matrix
    rot = cursor_mat.decompose()[1]
    x = rot @ Quaternion((0.0, 1.0, 0.0), radians(90)) 
    y = rot @ Quaternion((1.0, 0.0, 0.0), radians(-90))
    z = rot @ Quaternion((0.0, 0.0, 1.0), radians(-90))
    return x, y, z


def get_parent_orient():
    obj = bpy.context.object
    parent = obj.parent
    if parent:
        rot = parent.matrix_world.decompose()[1]
    else:
        rot = Quaternion()
    x = rot @ Quaternion((0.0, 1.0, 0.0), radians(90))
    y = rot @ Quaternion((1.0, 0.0, 0.0), radians(-90))
    z = rot @ Quaternion((0.0, 0.0, 1.0), radians(-90))
    return x, y, z


def get_custom_orient(matrix=None):
    if matrix:
        rot = matrix.decompose()[1]
    else:
        custom_mat = bpy.context.scene.transform_orientation_slots[0].custom_orientation.matrix.to_4x4()
        rot = custom_mat.decompose()[1]
    x = rot @ Quaternion((0.0, 1.0, 0.0), radians(90)) 
    y = rot @ Quaternion((1.0, 0.0, 0.0), radians(-90))
    z = rot @ Quaternion((0.0, 0.0, 1.0), radians(-90))
    return x, y, z



# --- VIEW ORIENT
def view_orient_matrix( orig_loc, use_standard = False ):
    context = bpy.context

    if context.mode == 'SCULPT': # TODO FIXME
        GP_col = context.object.GP_scene_Col
        l = GP_col.custom_position
    else:
        l = orig_loc
        
    if use_standard:
        rot = Quaternion( (0.0, 0.0, 1.0), radians(-90) )
    else:
        rot = context.region_data.view_matrix.inverted().decompose()[1]

    s = Vector((1.0, 1.0, 1.0))

    # --- VIEW ROTATE
    vm = Matrix.LocRotScale(l, rot, s) @ Matrix.Rotation(radians(180), 4, 'Z') @ Matrix.Rotation(radians(45), 4, 'Z')
    vm.normalized()

    # --- VIEW SCALE
    vms = Matrix.LocRotScale(l, rot, s) @ Matrix.Rotation(radians(90), 4, 'Y') @ Matrix.Rotation(radians(45), 4, 'X') @ Matrix.Rotation(radians(45), 4, 'Z')
    vms.normalized()
    
    
    return vm, vms