import bpy, gpu
from math import sin, cos, pi, copysign, radians, sqrt, ceil, acos
from gpu.types import (
        GPUBatch,
        GPUIndexBuf,
        GPUVertBuf,
        GPUVertFormat,
    )
from mathutils import Vector
from gpu import shader as sh
from gpu import matrix as mat
from ..shaders import (
    GPU_SHADER_3D_LINE_DASHED_UNIFORM_COLOR,
)



# --- NEW
def get_circle(position, radius, segments=60):
    if len(position) == 2:
        mx, my = position
        return [(radius * cos(i * 2 * pi / segments) + mx,
                 radius * sin(i * 2 * pi / segments) + my)
                for i in range(segments + 1)]
    elif len(position) == 3:
        mx, my, mz = position
        return [(radius * cos(i * 2 * pi / segments) + mx,
                 radius * sin(i * 2 * pi / segments) + my,
                 mz)
                for i in range(segments + 1)]
    else:
        raise ValueError("Position must be a 2D or 3D Vector")


def get_rectangle(position, width, height):
    if len(position) == 2:
        mx, my = position
        half_width = width / 2
        half_height = height / 2
        points = [
            (mx - half_width, my - half_height),
            (mx + half_width, my - half_height),
            (mx + half_width, my + half_height),
            (mx - half_width, my + half_height)
        ]
        return points

    elif len(position) == 3:
        mx, my, mz = position
        half_width = width / 2
        half_height = height / 2
        points = [
            (mx - half_width, my - half_height, mz),
            (mx + half_width, my - half_height, mz),
            (mx + half_width, my + half_height, mz),
            (mx - half_width, my + half_height, mz)
        ]
        return points

    else:
        raise ValueError("Position must be a 2D or 3D Vector")


def get_rectangle_with_bevels(position, width, height, bevel_size):
    if len(position) == 2:
        mx, my = position
        half_width = width / 2
        half_height = height / 2

        points = [
            (mx - half_width + bevel_size, my - half_height),
            (mx + half_width - bevel_size, my - half_height),
            (mx + half_width, my - half_height + bevel_size),
            (mx + half_width, my + half_height - bevel_size),
            (mx + half_width - bevel_size, my + half_height),
            (mx - half_width + bevel_size, my + half_height),
            (mx - half_width, my + half_height - bevel_size),
            (mx - half_width, my - half_height + bevel_size)
        ]
        return points

    elif len(position) == 3:
        mx, my, mz = position
        half_width = width / 2
        half_height = height / 2

        points = [
            (mx - half_width + bevel_size, my - half_height, mz),
            (mx + half_width - bevel_size, my - half_height, mz),
            (mx + half_width, my - half_height + bevel_size, mz),
            (mx + half_width, my + half_height - bevel_size, mz),
            (mx + half_width - bevel_size, my + half_height, mz),
            (mx - half_width + bevel_size, my + half_height, mz),
            (mx - half_width, my + half_height - bevel_size, mz),
            (mx - half_width, my - half_height + bevel_size, mz)
        ]
        return points

    else:
        raise ValueError("Position must be a 2D or 3D Vector")


def get_rounded_rectangle(position, width, height, radius, segments=5):
    if len(position) not in [2, 3]:
        raise ValueError("Position must be a 2D or 3D Vector")

    def add_rounded_corner(center, start_angle, end_angle, radius, segments):
        return [
            (
                center[0] + radius * cos(radians(angle)),
                center[1] + radius * sin(radians(angle))
            ) + center[2:] for angle in range(start_angle, end_angle, int((end_angle - start_angle) / segments))
        ]

    mx, my, *mz = position
    half_width = width / 2
    half_height = height / 2

    # Define corner centers
    corners = [
        (mx - half_width + radius, my - half_height + radius),  # Bottom left
        (mx + half_width - radius, my - half_height + radius),  # Bottom right
        (mx + half_width - radius, my + half_height - radius),  # Top right
        (mx - half_width + radius, my + half_height - radius)   # Top left
    ]

    points = []
    # Bottom left corner
    points += add_rounded_corner(corners[0], 180, 270, radius, segments)
    # Bottom edge
    points.append((mx + half_width - radius, my - half_height) + tuple(mz))
    # Bottom right corner
    points += add_rounded_corner(corners[1], 270, 360, radius, segments)
    # Right edge
    points.append((mx + half_width, my + half_height - radius) + tuple(mz))
    # Top right corner
    points += add_rounded_corner(corners[2], 0, 90, radius, segments)
    # Top edge
    points.append((mx - half_width + radius, my + half_height) + tuple(mz))
    # Top left corner
    points += add_rounded_corner(corners[3], 90, 180, radius, segments)
    # Left edge
    points.append((mx - half_width, my - half_height + radius) + tuple(mz))

    return points


def get_triangle_sliding(position, size, slide):
    if len(position) == 2:
        half_size = size / 2
        p3_x = position[0]-half_size + ((position[0]+half_size) - (position[0]-half_size)) * slide
        p3_y = position[1]+half_size + ((position[1]-half_size) - (position[1]+half_size)) * slide
        p = [
            (position[0]-half_size, position[1]+half_size), # p1
            (position[0]+half_size, position[1]+half_size), # p2
            (position[0]-half_size, position[1]-half_size), # p3
            (position[0]-half_size, p3_y), # p4
            (p3_x, position[1]+half_size), # p5
        ]
        lines = [p[0], p[1], p[0], p[2], p[3], p[4]]
        tri = [p[0], p[3], p[4]]
        return lines, tri

    else:
        z = position[2]
        half_size = size / 2
        p3_x = position[0]-half_size + ((position[0]+half_size) - (position[0]-half_size)) * slide
        p3_y = position[1]+half_size + ((position[1]-half_size) - (position[1]+half_size)) * slide
        p = [
            (position[0]-half_size, position[1]+half_size, z), # p1
            (position[0]+half_size, position[1]+half_size, z), # p2
            (position[0]-half_size, position[1]-half_size, z), # p3
            (position[0]-half_size, p3_y, z), # p4
            (p3_x, position[1]+half_size, z), # p5
        ]
        lines = [p[0], p[1], p[0], p[2], p[3], p[4]]
        tri = [p[0], p[3], p[4]]
        return lines, tri


def get_capsule(position, l, r, segments=60, axis='X'):
    if len(position) not in [2, 3]:
        raise ValueError("Position must be a 2D or 3D Vector")

    px = position[0]
    py = position[1]
    if len(position)==3:
        pz = position[2]

    points = []
    
    if axis == 'X':
        for i in range((segments // 2) + 1): # левый полукруг ?
            theta = pi * i / (segments / 2)
            x = px + r * cos(theta) 
            y = py + l / 2 + r * sin(theta)
            if len(position)==2:
                points.append((x, y))
            else:
                points.append((x, y, pz))

        for i in range(segments // 2, segments+1): # правый полукруг ?
            theta = pi * i / (segments / 2)
            x = px + r * cos(theta) 
            y = py - l / 2 + r * sin(theta)
            if len(position)==2:
                points.append((x, y))
            else:
                points.append((x, y, pz))

    elif axis == 'Y':
        for i in range((segments // 2) + 1): # нижний полукруг ?
            theta = pi * i / (segments / 2)
            x = px + l / 2 + r * sin(theta)
            y = py + r * cos(theta)
            if len(position)==2:
                points.append((x, y))
            else:
                points.append((x, y, pz))

        for i in range(segments // 2, segments+1): # верхний полукруг ?
            theta = pi * i / (segments / 2)
            x = px - l / 2 + r * sin(theta)
            y = py + r * cos(theta)
            if len(position)==2:
                points.append((x, y))
            else:
                points.append((x, y, pz))

    elif axis == 'Z':
        for i in range((segments // 2) + 1): # нижний полукруг ?
            theta = pi * i / (segments / 2)
            x = px + r * cos(theta) 
            z = pz + l / 2 + r * sin(theta)
            if len(position)==2:
                points.append((x, z))
            else:
                points.append((x, py, z))

        for i in range(segments // 2, segments+1): # верхний полукруг ?
            theta = pi * i / (segments / 2)
            x = px + r * cos(theta) 
            z = pz - l / 2 + r * sin(theta)
            if len(position)==2:
                points.append((x, z))
            else:
                points.append((x, py, z))


    return points



















def division(position, out, iner, segments=60):
    mx = position[0]
    my = position[1]

    prefs = bpy.context.preferences.system
    outRad_ = out * (prefs.dpi * prefs.pixel_size / 72)
    inerRad_ = iner * (prefs.dpi * prefs.pixel_size / 72)

    verts = []
    if len(position) == 3:
        for i in range(segments + 1):
            grad = (360 * i / segments) * pi / 180
            verts.append((outRad_ * cos(grad) + mx, outRad_ * sin(grad) + my, 0))
            verts.append((inerRad_ * cos(grad) + mx, inerRad_ * sin(grad) + my, 0))
    else:
        for i in range(segments + 1):
            grad = (360 * i / segments) * pi / 180
            verts.append((outRad_ * cos(grad) + mx, outRad_ * sin(grad) + my))
            verts.append((inerRad_ * cos(grad) + mx, inerRad_ * sin(grad) + my))

    return verts











# --- Supercircle

def distanceOfTwoPoints(px, py, qx, qy): # Supercircle distance
    n = 4 #change n for different tests (0.5 and 4)
    distance = (abs(qx - px)**n + abs(qy - py)**n)**(1/n)
    return distance

def squircle( position, rx = 1.0, ry = 1.0, n = 0.4, segments = 60 ):

    sign = lambda x: copysign(1, x) # TODO вынести за пределы функции

    dim = len(position)

    mx = position[0]
    my = position[1]

    if dim == 2:
        mz = 0.0
    else:
        mz = position[2]


    mul = 2 * pi / segments
    #mul = (1.0 / (segments - 1)) * (pi * 2)
    if dim == 2:
        verts = [(
                rx * abs( cos( i * mul ) )**n * sign( cos( i * mul ) ) + mx,
                ry * abs( sin( i * mul ) )**n * sign( sin( i * mul ) ) + my,
                )
                for i in range(segments + 1)]
    else:
        verts = [(
                rx * abs( cos( i * mul ) )**n * sign( cos( i * mul ) ) + mx,
                ry * abs( sin( i * mul ) )**n * sign( sin( i * mul ) ) + my,
                mz,
                )
                for i in range(segments + 1)]

    return verts


# --- Arc
def arc( position, radius = 1, a_angle = 0.0, b_angle = 45.0, points = 4 ):
    
    dim = len(position)

    mx = position[0]
    my = position[1]
    if dim == 2:
        prefs = bpy.context.preferences.system
        r = radius * (prefs.dpi * prefs.pixel_size / 72)
        mz = 0.0
    else:
        r = radius
        mz = position[2]
    
    
    startangle = radians(a_angle)
    endangle = radians(b_angle)
    a = (endangle - startangle) / points

    if dim == 2:
        verts = [(
            r * cos( i * a + startangle) + mx,
            r * sin( i * a + startangle) + my,
            )
            for i in range(points) ]
    else:
        verts = [(
                r * cos( i * a + startangle) + mx,
                r * sin( i * a + startangle) + my,
                mz,
                )
                for i in range(points) ]


    return verts












def capsule_2d(position=(0.0,0.0), size=2.0, radius=1.0, segments=60, align='HORIZONTAL', dpi=False, closing_point=False):
    mx = position[0]
    my = position[1]
    
    if dpi:
        prefs = bpy.context.preferences.system
        radius = radius * (prefs.dpi * prefs.pixel_size / 72)

    if segments is None:
        max_pixel_error = 0.25  # TODO: multiply 0.5 by display dpi
        segments = int(ceil(pi / acos(1.0 - max_pixel_error / radius)))
        segments = max(segments, 8)
        segments = min(segments, 1000)
    
    half_size = size / 2
    mul = (1.0 / (segments - 1)) * (pi * 2)
    verts = []
    for i in range(segments):
        x = radius * sin(i * mul) + mx
        y = radius * cos(i * mul) + my
        
        if align == 'HORIZONTAL':
            if x < mx:
                x -= half_size
            else:
                x += half_size
        elif align == 'VERTICAL':
            if y < my:
                y -= half_size
            else:
                y += half_size
        
        verts.append((x, y))
    
    if closing_point:
        verts.append( verts[0] )

    return verts



def sector(position = (0.0, 0.0), segments=6, radius=1.0, startangle=0.0, endangle=90.0):
    mx = position[0]
    my = position[1]

    newpoints = []

    startangle = radians(startangle)
    endangle = radians(endangle)
    segments += 1

    angle = (endangle - startangle) / segments
    x = cos(startangle) * radius + mx
    y = sin(startangle) * radius + my
    newpoints.append([x, y, 0])
    j = 1
    while j < segments:
        t = angle * j
        x = cos(t + startangle) * radius + mx
        y = sin(t + startangle) * radius + my
        newpoints.append([x, y, 0])
        j += 1
    x = cos(endangle) * radius + mx
    y = sin(endangle) * radius + my
    newpoints.append([x, y, 0])

    return newpoints



def ring_2d(position, out, iner, segments=60):
    mx = position[0]
    my = position[1]

    ring1 = []
    ring2 = []
    verts = []
    faces = []
    
    #prefs = bpy.context.preferences.system
    outRad_ = out #* (prefs.dpi * prefs.pixel_size / 72)
    inerRad_ = iner #* (prefs.dpi * prefs.pixel_size / 72)

    for i in range(segments+1):
        grad = (360 * i / segments) * pi / 180
        ring1.append([outRad_ * cos(grad) + mx, outRad_ * sin(grad) + my])
        ring2.append([inerRad_ * cos(grad) + mx, inerRad_ * sin(grad) + my])

        faces.append([i, i+1, i+1+segments])
        faces.append([i, i+segments, i+1+segments])
           
    verts.extend(ring1)
    verts.extend(ring2)
    
    verts_new = []

    for f in faces:
        for i in f:
            verts_new.append(verts[i])
    return verts_new



def cross_3d(x=1.0, y=1.0, l=0.2, axis='Z'):
    if axis == 'X':
        co = [
            (0.0, x, 0.0), (0.0, l, 0.0), 
            (0.0, -l, 0.0), (0.0, -x, 0.0),
            (0.0, 0.0, y), (0.0, 0.0, l),
            (0.0, 0.0, -l), (0.0, 0.0, -y),
            ]

    elif axis == 'Y':
        co = [
            (x, 0.0, 0.0), (l, 0.0, 0.0), 
            (-l, 0.0, 0.0), (-x, 0.0, 0.0),
            (0.0, 0.0, y), (0.0, 0.0, l),
            (0.0, 0.0, -l), (0.0, 0.0, -y),
            ]

    elif axis == 'Z':
        co = [
            (x, 0.0, 0.0), (l, 0.0, 0.0), 
            (-l, 0.0, 0.0), (-x, 0.0, 0.0),
            (0.0, y, 0.0), (0.0, l, 0.0),
            (0.0, -l, 0.0), (0.0, -y, 0.0),
            ]
    
    return co



def circle_pie_2d(position, radius, angle=90, segments=60):
    mx = position[0]
    my = position[1]

    prefs = bpy.context.preferences.system
    radius = radius * (prefs.dpi * prefs.pixel_size / 72)
    
    seg = segments + 1
    end = seg - 1
    vertices = []
    for i in range(seg):
        grad = (angle * i / end) * pi / 180
        vertices.append([radius * cos(grad) + mx, radius * sin(grad) + my])
    
    """ vertices = [(radius * cos( angle * i / segments) + mx,
                 radius * sin( angle * i / segments) + my)
                 for i in range(segments + 1)] """

    return vertices



def pie_2d(position, out, iner, angle=90, segments=60):
    mx = position[0]
    my = position[1]

    ring1 = []
    ring2 = []
    verts = []
    faces = []
    
    #prefs = bpy.context.preferences.system
    #dpi = bpy.context.preferences.system.ui_scale
    outRad_ = out #* dpi #* (prefs.dpi * prefs.pixel_size / 72)
    inerRad_ = iner #* dpi  #* (prefs.dpi * prefs.pixel_size / 72)

    seg = segments + 1
    end = seg - 1
    for i in range(seg):
        grad = (angle * i / end) * pi / 180
        ring1.append([outRad_ * cos(grad) + mx, outRad_ * sin(grad) + my, 0.0])
        ring2.append([inerRad_ * cos(grad) + mx, inerRad_ * sin(grad) + my, 0.0])

        if i != end:
            faces.append([i, i+1, i+1+seg])
            faces.append([i, i+seg, i+1+seg])
    
    verts.extend(ring1)
    verts.extend(ring2)
    
    # --- Get Vertices Loop
    verts_loop = []
    verts_loop.extend(ring1)
    ring2.reverse()
    verts_loop.extend(ring2)
    verts_loop.append(verts_loop[0]) # add the first point to the end to close

    verts_new = []
    for f in faces:
        for i in f:
            verts_new.append(verts[i])
    return verts_new, verts_loop



def draw_circle_2d(position, color, radius, segments=60, type='LINE_STRIP'):
    with mat.push_pop():
        mat.translate(position)
        mat.scale_uniform(radius)
        mul = (1.0 / (segments - 1)) * (pi * 2)
        verts = [(sin(i * mul), cos(i * mul)) for i in range(segments)]
        fmt = GPUVertFormat()
        pos_id = fmt.attr_add(id='pos', comp_type='F32', len=2, fetch_mode='FLOAT')
        vbo = GPUVertBuf(len=len(verts), format=fmt)
        vbo.attr_fill(id=pos_id, data=verts)
        batch = GPUBatch(type=type, buf=vbo)
        shader = sh.from_builtin('UNIFORM_COLOR')
        batch.program_set(shader)
        shader.uniform_float('color', color)
        batch.draw()



def draw_ring_2d(position, color, scale=1.0, out=1.0, iner=0.5, segments=60, type='TRIS'):
    ring = mat.push_pop()
    with ring:
        mat.translate(position)
        mat.scale_uniform(scale)
        verts = ring_2d((0,0), out, iner, segments)
        fmt = GPUVertFormat()
        pos_id = fmt.attr_add(id='pos', comp_type='F32', len=2, fetch_mode='FLOAT')
        vbo = GPUVertBuf(len=len(verts), format=fmt)
        vbo.attr_fill(id=pos_id, data=verts)
        batch = GPUBatch(type=type, buf=vbo)
        shader = sh.from_builtin('UNIFORM_COLOR')
        batch.program_set(shader)
        shader.uniform_float('color', color)
        batch.draw()



def draw_3d_dashed(verts, color, type='TRIS', colors_len=0, dash_width=10.0, dash_factor=0.5):
    dim = len(verts[0])
    fmt = GPUVertFormat()
    pos_id = fmt.attr_add(id='pos', comp_type='F32', len=dim, fetch_mode='FLOAT')
    vbo = GPUVertBuf(len=len(verts), format=fmt)
    vbo.attr_fill(id=pos_id, data=verts)
    batch = GPUBatch(type=type, buf=vbo)

    shader = GPU_SHADER_3D_LINE_DASHED_UNIFORM_COLOR()
    batch.program_set(shader)
    shader.uniform_float('color', color)
    viewport_size = gpu.state.viewport_get()
    shader.uniform_float('viewport_size', (viewport_size[2], viewport_size[3]))
    shader.uniform_int('colors_len', colors_len)
    shader.uniform_float('dash_width', dash_width)
    shader.uniform_float('dash_factor', dash_factor)

    batch.draw()



def batch_for_shader(shader, type, content, *, indices=None): # --- ??
    for data in content.values():
        vbo_len = len(data)
        break
    else:
        raise ValueError("Empty 'content'")

    vbo_format = shader.format_calc()
    vbo = GPUVertBuf(vbo_format, vbo_len)

    for id, data in content.items():
        if len(data) != vbo_len:
            raise ValueError("Length mismatch for 'content' values")
        vbo.attr_fill(id, data)

    if indices is None:
        return GPUBatch(type=type, buf=vbo)
    else:
        ibo = GPUIndexBuf(type=type, seq=indices)
        return GPUBatch(type=type, buf=vbo, elem=ibo)



def draw_batch_mat(position, verts, color, scale=1.0, type='TRIS'):  # --- ??
    ring = mat.push_pop()
    with ring:
        dim = len(verts[0])

        mat.translate(position)
        mat.scale_uniform(scale)
        fmt = GPUVertFormat()
        pos_id = fmt.attr_add(id='pos', comp_type='F32', len=dim, fetch_mode='FLOAT')
        vbo = GPUVertBuf(len=len(verts), format=fmt)
        vbo.attr_fill(id=pos_id, data=verts)
        batch = GPUBatch(type=type, buf=vbo)
        shader = sh.from_builtin('UNIFORM_COLOR')
        batch.program_set(shader)
        shader.uniform_float('color', color)
        batch.draw()



def draw_batch(verts, color, type='TRIS', custom_shader=None):  # --- ??
    dim = len(verts[0])
    fmt = GPUVertFormat()
    pos_id = fmt.attr_add(id='pos', comp_type='F32', len=dim, fetch_mode='FLOAT')
    vbo = GPUVertBuf(len=len(verts), format=fmt)
    vbo.attr_fill(id=pos_id, data=verts)
    batch = GPUBatch(type=type, buf=vbo)
    if custom_shader:
        shader = custom_shader
    else:
        shader = sh.from_builtin('UNIFORM_COLOR')
    batch.program_set(shader)
    shader.uniform_float('color', color)
    batch.draw()



def gp_batch_for_shader( shader, type, content, *, indices = None ):
    """
    Return a batch already configured and compatible with the shader.

    :arg shader: shader for which a compatible format will be computed.
    :type shader: :class:`gpu.types.GPUShader`
    :arg type: "'POINTS', 'LINES', 'TRIS' or 'LINES_ADJ'".
    :type type: str
    :arg content: Maps the name of the shader attribute with the data to fill the vertex buffer.
    :type content: dict
    :return: compatible batch
    :rtype: :class:`gpu.types.Batch`
    """
    from gpu.types import (
        GPUBatch,
        GPUIndexBuf,
        GPUVertBuf,
        GPUVertFormat,
    )

    def recommended_comp_type(attr_type):
        if attr_type in {'FLOAT', 'VEC2', 'VEC3', 'VEC4', 'MAT3', 'MAT4'}:
            return 'F32'
        if attr_type in {'UINT', 'UVEC2', 'UVEC3', 'UVEC4'}:
            return 'U32'
        # `attr_type` in {'INT', 'IVEC2', 'IVEC3', 'IVEC4', 'BOOL'}.
        return 'I32'

    def recommended_attr_len(attr_name):
        attr_len = 1
        try:
            item = content[attr_name][0]
            while True:
                attr_len *= len(item)
                item = item[0]
        except (TypeError, IndexError):
            pass
        return attr_len

    def recommended_fetch_mode(comp_type):
        if comp_type == 'F32':
            return 'FLOAT'
        return 'INT'

    for data in content.values():
        vbo_len = len(data)
        break
    else:
        raise ValueError("Empty 'content'")

    vbo_format = GPUVertFormat()
    attrs_info = shader.attrs_info_get()
    for name, attr_type in attrs_info:
        comp_type = recommended_comp_type(attr_type)
        attr_len = recommended_attr_len(name)
        vbo_format.attr_add(id=name, comp_type=comp_type, len=attr_len, fetch_mode=recommended_fetch_mode(comp_type))

    vbo = GPUVertBuf(vbo_format, vbo_len)

    for id, data in content.items():
        if len(data) != vbo_len:
            raise ValueError("Length mismatch for 'content' values")
        vbo.attr_fill(id, data)

    if indices is None:
        return GPUBatch(type=type, buf=vbo)
    else:
        ibo = GPUIndexBuf(type=type, seq=indices)
        return GPUBatch(type=type, buf=vbo, elem=ibo)



def draw_custom_batch(self, verts, type='TRIS', matrix=None, select_id=None, color_fal=None):  # --- ??
    shader = sh.from_builtin('UNIFORM_COLOR')

    batch = batch_for_shader(shader, type, {'pos': verts})

    if matrix is None:
        matrix = self.matrix_world

    shader.bind()

    if select_id is not None:
        gpu.select.load_id(select_id)
    else:
        if self.is_highlight:
            color = (*self.color_highlight, self.alpha_highlight)
        else:
            if color_fal is None:
                color = (*self.color, self.alpha)
            else:
                color = color_fal
        shader.uniform_float('color', color)

    with gpu.matrix.push_pop():
        gpu.matrix.multiply_matrix(matrix)
        batch.draw(shader)

    

def draw_custom_batch_mw(verts, type='TRIS', matrix=None, select_id=None, color=None):  # --- ??
    shader = sh.from_builtin('UNIFORM_COLOR')

    batch = batch_for_shader(shader, type, {'pos': verts})
    
    shader.bind()

    if select_id is not None:
        gpu.select.load_id(select_id)
    else:
        shader.uniform_float('color', color)

    with gpu.matrix.push_pop():
        gpu.matrix.multiply_matrix(matrix)
        batch.draw(shader)

    
        
        



# --- Texture
def get_uv_coords(icon_id):
    if icon_id is None:
        return ((0, 0), (1, 0), (1, 1), (0, 1))
    row = (icon_id-1) // 10
    col = (icon_id-1) % 10
    x1 = col / 10
    y1 = row / 10
    x2 = (col + 1) / 10
    y2 = (row + 1) / 10
    return ((x1, y1), (x2, y1), (x2, y2), (x1, y2))




def draw_texture_2d(texture, position, width, height, icon_id=None):
    from gpu_extras.batch import batch_for_shader
    coords = get_uv_coords(icon_id)

    shader = gpu.shader.from_builtin('IMAGE')
    batch = batch_for_shader(
        shader, 'TRI_FAN',
        {"pos": coords, "texCoord": coords},
    )

    with gpu.matrix.push_pop():
        pos = (position[0], position[1])
        gpu.matrix.translate(pos)
        gpu.matrix.scale((width, height))

        shader = gpu.shader.from_builtin('IMAGE')

        if isinstance(texture, int):
            # Call the legacy bgl to not break the existing API
            import bgl
            bgl.glActiveTexture(bgl.GL_TEXTURE0)
            bgl.glBindTexture(bgl.GL_TEXTURE_2D, texture)
            shader.uniform_int("image", 0)
        else:
            shader.uniform_sampler("image", texture)

        batch.draw(shader)







def draw_vertex(coords, color):
    shader = sh.from_builtin('UNIFORM_COLOR')

    batch = batch_for_shader(shader, 'POINTS', {"pos": coords})
    shader.bind()
    shader.uniform_float("color", color)

    batch.draw(shader)
   

def draw_edge(coords, color, indices=None):
    if indices is None:
        indices=([0,1])

    shader = sh.from_builtin('UNIFORM_COLOR')
    batch = batch_for_shader(shader, 'LINES', {"pos": coords})
    shader.bind()
    shader.uniform_float("color", color)
    batch.draw(shader)


# --- DRAW PIE
def draw_pie_2d(position, outRad, inerRad, segments=60, pie=[20, 20, 20, 20]): # TODO
    mx = position[0]
    my = position[1]


    ring1 = []
    ring2 = []
    verts = []
    faces = []
    

    prefs = bpy.context.preferences.system
    outRad_ = outRad * (prefs.dpi * prefs.pixel_size / 72)
    inerRad_ = inerRad * (prefs.dpi * prefs.pixel_size / 72)

    for i in range(segments):
        grad = (360 * i / segments) * pi / 180
        ring1.append([outRad_ * cos(grad) + mx, outRad_ * sin(grad) + my])
        ring2.append([inerRad_ * cos(grad) + mx, inerRad_ * sin(grad) + my])

   
        end = segments - 1
        if i == end:
            faces.append([i, i+1, i+segments])
            faces.append([i, i+1, i-i]) 
        else:
            faces.append([i, i+1, i+1+segments])
            faces.append([i, i+segments, i+1+segments])
           
    verts.extend(ring1)
    verts.extend(ring2)
    
    verts_new = []

    #print(faces)
    for f in faces:
        for i in f:
            verts_new.append(verts[i])
    return verts_new
