import bpy
from bpy.types import Operator
from bpy.props import EnumProperty, FloatVectorProperty
from mathutils import Quaternion, Matrix
from math import radians
from ..func import pro_center_orient_reset
from ... import get_prefs



def add_object_align_init(context, operator): # TODO FIXME TEST
    '''
    Return a matrix using the operator settings and view context.
    :arg context: The context to use.
    :type context: :class:`bpy.types.Context`
    :arg operator: The operator, checked for location and rotation properties.
    :type operator: :class:`bpy.types.Operator`
    :return: the matrix from the context and settings.
    :rtype: :class:`mathutils.Matrix`
    '''

    from mathutils import Matrix, Vector, Euler
    properties = operator.properties if operator is not None else None

    space_data = context.space_data
    if space_data and space_data.type != 'VIEW_3D':
        space_data = None

    # location
    if operator and properties.is_property_set('location'):
        location = Matrix.Translation(Vector(properties.location))
    else:
        location = Matrix.Translation(context.scene.cursor.location)

        if operator:
            properties.location = location.to_translation()

    # rotation
    add_align_preference = context.preferences.edit.object_align
    if operator:
        if not properties.is_property_set('rotation'):
            # So one of 'align' and 'rotation' will be set
            properties.align = add_align_preference

        if properties.align == 'WORLD':
            rotation = properties.rotation.to_matrix().to_4x4()
        elif properties.align == 'VIEW':
            rotation = space_data.region_3d.view_matrix.to_3x3().inverted()
            rotation.resize_4x4()
            properties.rotation = rotation.to_euler()
        elif properties.align == 'CURSOR':
            rotation = context.scene.cursor.matrix
            rotation.col[3][0:3] = 0.0, 0.0, 0.0
            properties.rotation = rotation.to_euler()
        else:
            rotation = properties.rotation.to_matrix().to_4x4()
    else:
        if (add_align_preference == 'VIEW') and space_data:
            rotation = space_data.region_3d.view_matrix.to_3x3().inverted()
            rotation.resize_4x4()
        elif add_align_preference == 'CURSOR':
            rotation = context.scene.cursor.rotation_euler.to_matrix().to_4x4()
        else:
            rotation = Matrix()

    return location @ rotation







class GP_OT_cone_operator(Operator): 
    bl_idname = 'gp.cone_operator'
    bl_label = 'Translate'
    bl_description = 'Translate'
    bl_options = {'BLOCKING'}


    axis: EnumProperty(
        name='Axis',
        description='Axis',
        items=[
            ('X', 'X', '', '', 0),
            ('Y', 'Y', '', '', 1),
            ('Z', 'Z', '', '', 2),
            ('VIEW', 'View', '', '', 3),
            ],
        default='VIEW',
        )

    matrix_basis: FloatVectorProperty(size=(4, 4), subtype='MATRIX')
    pp = None
    axis_set = None
    

    @staticmethod
    def get_orient_matrix(self):
        mat = self.matrix_basis.copy().to_3x3()
        if self.axis == 'X':
            mat = mat @ Quaternion( (0.0, 1.0, 0.0), radians(90) ).to_matrix().to_3x3()
        elif self.axis == 'Y':
            mat = mat @ Quaternion( (1.0, 0.0, 0.0), radians(90) ).to_matrix().to_3x3()
        return mat

    
    def execute(self, context):
        tool_settings = context.scene.tool_settings
        tpp = context.scene.tool_settings.transform_pivot_point
        settings = context.scene.GP_scene_set
        orient_slots = context.window.scene.transform_orientation_slots[0].type
        



        # --- если включен VR то включаем снап при скейле
        use_snap_translate = tool_settings.use_snap_translate
        use_snap_rotate = tool_settings.use_snap_rotate
        use_snap_scale = tool_settings.use_snap_scale
        snap_target = tool_settings.snap_target
        if settings.gzm_type == 'VR':
            tool_settings.use_snap_translate = True
            tool_settings.use_snap_rotate = True
            tool_settings.use_snap_scale = True
            tool_settings.snap_target = 'CENTER'

        if (tpp or settings.position) == 'INDIVIDUAL_ORIGINS':
            bpy.ops.transform.transform(
                'INVOKE_DEFAULT',
                mode = 'TRANSLATION',
                #orient_axis = self.axis,
                orient_matrix = self.get_orient_matrix(self),
                constraint_axis = self.axis_set,
                release_confirm = True,
                )
        else:
            if orient_slots == 'NORMAL' or settings.orientation != 'NONE': 
                bpy.ops.transform.transform(
                    'INVOKE_DEFAULT',
                    mode = 'TRANSLATION',
                    #orient_axis = self.axis,
                    orient_matrix = self.get_orient_matrix(self),
                    constraint_axis = self.axis_set,
                    release_confirm = True,
                    center_override = self.pp,
                    )
            else: # --- для избежания неточности при перемещении
                bpy.ops.transform.transform(
                    'INVOKE_DEFAULT',
                    mode = 'TRANSLATION',
                    orient_axis = self.axis,
                    constraint_axis = self.axis_set,
                    release_confirm = True,
                    center_override = self.pp,
                    )

        if settings.gzm_type == 'VR': # --- если включен VR то включаем снап при скейле
            tool_settings.use_snap_translate = use_snap_translate
            tool_settings.use_snap_rotate = use_snap_rotate
            tool_settings.use_snap_scale = use_snap_scale
            tool_settings.snap_target = snap_target

        pro_center_orient_reset(context) # --- Temporary
        return {'FINISHED'}

    def modal(self, context, event):
        if event.value == 'RELEASE' or event.value_prev == 'RELEASE' or event.type_prev in {'RET', 'LEFTMOUSE'} or event.type in {'RET', 'LEFTMOUSE'}:
            return {'FINISHED'}
        elif event.value in {'ESC', 'RIGHTMOUSE'} or event.value_prev in {'ESC', 'RIGHTMOUSE'}:
            return {'CANCELLED'}
        return {'RUNNING_MODAL'}


    def invoke(self, context, event):
        props = get_prefs()
        #settings = context.scene.GP_scene_set

        if self.pp is None:
            self.pp = self.matrix_basis.translation

        if self.axis == 'X':
            self.axis_set = (True, False, False)
        elif self.axis == 'Y':
            self.axis_set = (False, True, False)
        elif self.axis == 'Z':
            self.axis_set = (False, False, True)
        else:
            self.axis_set = (True, True, True)


        # --- Transform From Value
        if event.type == 'MIDDLEMOUSE' and event.value == 'CLICK':
            bpy.ops.gp.transform_from_value(
                'INVOKE_DEFAULT',
                axis = self.axis_set,
                operator = 'TRANSLATION',
                pp = self.pp,
                matrix_basis = self.get_orient_matrix(self),
                )
            return {'FINISHED'}


        # --- PT
        if event.type == 'MIDDLEMOUSE': # (event.ctrl and event.shift) or 
            bpy.ops.gp.pivot_transform(
                'INVOKE_DEFAULT',
                TRANSFORM_OT_transform = {
                    'mode': 'TRANSLATION',
                    'orient_matrix': self.get_orient_matrix(self),
                    'constraint_axis': self.axis_set,
                    'release_confirm': True,
                    'center_override': self.pp,
                    },
                )


        elif context.mode == 'OBJECT':
            if (event.ctrl and props.obCone_ctrl == 'DUPLICATE') or (event.shift and props.obCone_shift == 'DUPLICATE'):
                bpy.ops.object.duplicate_move(
                    'INVOKE_DEFAULT',
                    TRANSFORM_OT_translate = {
                        'orient_matrix': self.get_orient_matrix(self),
                        'constraint_axis': self.axis_set,
                        'release_confirm': True,
                        },
                    )
            elif (event.ctrl and props.obCone_ctrl == 'DUPLICATE_LINKED') or (event.shift and props.obCone_shift == 'DUPLICATE_LINKED'):
                bpy.ops.object.duplicate_move_linked(
                    'INVOKE_DEFAULT',
                    OBJECT_OT_duplicate = {
                        'linked': True,
                        'mode': 'TRANSLATION',
                        },
                    TRANSFORM_OT_translate = {
                        'orient_matrix': self.get_orient_matrix(self),
                        'constraint_axis': self.axis_set,
                        'release_confirm': True,
                        },
                    )
            else:
                return self.execute(context)
                #self.translate( self, context, self.axis_set, self.get_orient_matrix(self) )


        elif context.mode == 'EDIT_MESH':
            if (event.ctrl and props.meCone_ctrl == 'INSET') or (event.shift and props.meCone_shift == 'INSET'):
                bpy.ops.mesh.inset('INVOKE_DEFAULT', release_confirm = True)

            elif (event.ctrl and props.meCone_ctrl == 'EXTRUDE') or (event.shift and props.meCone_shift == 'EXTRUDE'):
                bpy.ops.mesh.extrude_context_move(
                    'INVOKE_DEFAULT',
                    MESH_OT_extrude_context = {
                        'use_normal_flip': False,
                        'mirror': False,
                        },
                    TRANSFORM_OT_translate = {
                        'orient_matrix': self.get_orient_matrix(self),
                        'constraint_axis': self.axis_set,
                        'release_confirm':True,
                        },
                    )

            elif (event.ctrl and props.meCone_ctrl == 'EXTRUDE_MANIFLOD') or (event.shift and props.meCone_shift == 'EXTRUDE_MANIFLOD'):
                bpy.ops.mesh.extrude_manifold(
                    'INVOKE_DEFAULT', 
                    MESH_OT_extrude_region = {
                        'use_normal_flip': False,
                        'use_dissolve_ortho_edges': True,
                        'mirror': False,
                        },
                    TRANSFORM_OT_translate = {
                        'orient_matrix': self.get_orient_matrix(self),
                        'constraint_axis': self.axis_set,
                        'mirror': False, 
                        'release_confirm': True,
                        },
                    )

            else:
                return self.execute(context)
                #self.translate( self, context, axis_set, self.get_orient_matrix(self) )

        # elif context.mode in {'EDIT_CURVE', 'EDIT_SURFACE'}:
        # elif context.mode in {'EDIT_ARMATURE', 'POSE'}
        else:
            return self.execute(context)

      
        return {'RUNNING_MODAL'}  


    
classes = [
    GP_OT_cone_operator,
    ]


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)