import bpy
from math import radians, degrees
from bpy.props import (
    FloatProperty,
    IntProperty,
    BoolProperty,
    EnumProperty,
)



class GP_OT_radial_array(bpy.types.Operator):
    bl_idname = 'gp.radial_array'
    bl_label = 'Radial Array'
    bl_description = 'Create a radial array of selected objects'
    bl_options = {'REGISTER', 'UNDO'}


    axis: EnumProperty(
        name='Axis',
        description='Axis to rotate around',
        items=[
            ('X', 'X', 'Rotate around X axis'),
            ('Y', 'Y', 'Rotate around Y axis'),
            ('Z', 'Z', 'Rotate around Z axis')
        ],
        default='X',
    )
    
    degrees: FloatProperty(
        name='Degrees',
        description='Angle in degrees',
        default=radians(360.0),
        min=0.0,
        max=360.0,
        subtype='ANGLE',
    )

    number: IntProperty(
        name='Number',
        description='Number of instances',
        default=6,
        min=1,
    )
    
    merge_objects: BoolProperty(
        name='Merge Objects',
        description='Merge all new objects into one',
        default=False,
    )
    
    
    def execute(self, context):
        active_object = context.active_object

        if active_object is None:
            self.report({'ERROR'}, 'No active object')
            return {'CANCELLED'}
        
        new_objects = []
        
        if round(degrees(self.degrees)) == 360.0:
            _range = range(self.number)
            _radians_step = self.degrees / self.number
            _last_pass = True
        else:
            _range = range(self.number-1)
            _radians_step = self.degrees / (self.number-1)
            _last_pass = False
            
        for i in _range:
            if _last_pass and i==(self.number-1):
                continue
            angle = _radians_step * (i+1)
            bpy.ops.object.select_all(action='DESELECT')
            active_object.select_set(True)
            context.view_layer.objects.active = active_object
            bpy.ops.object.duplicate(linked=False)
            bpy.ops.transform.rotate(value=angle, orient_axis=self.axis)
            new_objects.append(context.active_object)
        
        if self.merge_objects:
            bpy.ops.object.select_all(action='DESELECT')
            for obj in new_objects:
                obj.select_set(True)
            active_object.select_set(True)
            context.view_layer.objects.active = active_object
            bpy.ops.object.join()
        return {'FINISHED'}



classes = [
    GP_OT_radial_array,
    ]


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls) 