import bpy
import bmesh

from mathutils import Matrix, Vector
from math import radians, pi



def main(context, angle, steps, offset):
    obj = context.active_object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    
    
    
    axis = Vector((0.0, 0.0, 1.0))
    angle_step = -angle / steps
    
    selected_verts = [v for v in bm.verts if v.select]
    selected_edges = [e for e in bm.edges if e.select]
    selected_faces = [f for f in bm.faces if f.select]
    
    for i in range(steps):
        # Extrude selection
        data = bmesh.ops.extrude_face_region(bm, geom=selected_faces)
        #print(data['geom'])
        verts_new = [v for v in data['geom'] if isinstance(v, bmesh.types.BMVert)]
        selected_faces = [f for f in data['geom'] if isinstance(f, bmesh.types.BMFace)]
        
        #bmesh.ops.extrude_edge_only(bm, edges=selected_edges) #, use_select_history=True
        # Rotate extruded geometry around pivot
        M = Matrix.Translation(Vector( (0.0, 0.0, offset/steps ) ) ) @ Matrix.Rotation(angle_step, 4, axis)
        bmesh.ops.transform(bm, matrix=M, verts=verts_new)
        # Merge extruded geometry into mesh
        #bmesh.ops.delete(bm, geom=[f for f in bm.faces if f.select])
    
    
    bmesh.update_edit_mesh(me)
    bpy.ops.ed.undo_push(message='Revolve From Gizmo*')
    
    
class RevolveOperator(bpy.types.Operator):
    bl_idname = "mesh.revolve_operator"
    bl_label = "Spin Operator" 
    bl_options = {'REGISTER', 'UNDO'} # , 'BLOCKING', 'UNDO_GROUPED'
    
    
    angle: bpy.props.FloatProperty(
        name="Angle",
        description="Angle to spin around",
        subtype='ANGLE',
        default=radians(10),
    )
    
    steps: bpy.props.IntProperty(
        name="Steps",
        description="Number of steps for the spin",
        default=2,
        min=1,
        soft_max=300
    )
    
    offset: bpy.props.FloatProperty(
        name="Offset",
        description="",
        subtype='DISTANCE',
        default=0,
    )
    
    
    def execute(self, context):
        main(context, self.angle, self.steps, self.offset)
        return {'FINISHED'}
    
        
    def invoke(self, context, event):
        self.execute(context)

        if context.space_data.type == 'VIEW_3D':
            wm = context.window_manager
            wm.gizmo_group_type_ensure(GP_GGT_revolve_gizmo.bl_idname)

        return {'FINISHED'}



class GP_GGT_revolve_gizmo(bpy.types.GizmoGroup):
    bl_idname = "GP_GGT_revolve_gizmo"
    bl_label = "Revolve"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW'
    bl_options = {'3D', 'EXCLUDE_MODAL'}

    # Helper functions
    @staticmethod
    def my_target_operator(context):
        wm = context.window_manager
        op = wm.operators[-1] if wm.operators else None
        if isinstance(op, RevolveOperator):
            return op
        return None

    @staticmethod
    def my_view_orientation(context):
        rv3d = context.space_data.region_3d
        view_inv = rv3d.view_matrix.to_3x3()
        return view_inv.normalized()

    @classmethod
    def poll(cls, context):
        op = cls.my_target_operator(context)
        if op is None:
            wm = context.window_manager
            wm.gizmo_group_type_unlink_delayed(GP_GGT_revolve_gizmo.bl_idname)
            return False
        return True

    def setup(self, context):
        from mathutils import Matrix, Vector


        def move_get_cb():
            op = GP_GGT_revolve_gizmo.my_target_operator(context)
            return op.offset

        def move_set_cb(value):
            op = GP_GGT_revolve_gizmo.my_target_operator(context)
            bpy.context.active_operator.offset = value
            
            
            
            #bpy.context.window_manager.print_undo_steps()
            # XXX, this may change!
            #bpy.context.active_operator.bl_rna.repeated
            #bpy.context.active_operator.execute(context)
            #op.offset = value
            op.execute(context)
            #op.update(context)

        gz = self.gizmos.new("GIZMO_GT_arrow_3d")
        
        #op = GP_GGT_revolve_gizmo.my_target_operator(context)
        gz.target_set_handler("offset", get=move_get_cb, set=move_set_cb)
        #gz.target_get_range(target)
        #gz.target_set_prop("offset", op, 'offset')
        gz.use_draw_value = True

        gz.color = 0.8, 0.8, 0.8
        gz.alpha = 0.5

        gz.color_highlight = 1.0, 1.0, 1.0
        gz.alpha_highlight = 1.0

        gz.scale_basis = 1.0
        
        
        
        
        
        # ----
        # Dial

        def direction_get_cb():
            op = GP_GGT_revolve_gizmo.my_target_operator(context)
            return op.angle

        def direction_set_cb(value):
            op = GP_GGT_revolve_gizmo.my_target_operator(context)
            op.angle = value
            # XXX, this may change!
            bpy.ops.ed.undo()
            op.execute(context)
            #op.update(context)

        gz = self.gizmos.new("GIZMO_GT_dial_3d")
        gz.target_set_handler("offset", get=direction_get_cb, set=direction_set_cb)
        gz.draw_options = {'ANGLE_VALUE'}
        
        
        gz.wrap_angle = False
        gz.arc_inner_factor = 0.9
        gz.click_value = pi * 2
        gz.use_draw_value = True
        
        gz.scale_basis = 2.0
        gz.line_width = 1.5
        
        gz.color = 0.8, 0.8, 0.8
        gz.alpha = 0.5

        gz.color_highlight = 1.0, 1.0, 1.0
        gz.alpha_highlight = 1.0

        self.gizmo_dial = gz
        
        

    def draw_prepare(self, context):
        from mathutils import Vector

        view_inv = self.my_view_orientation(context)

        #self.view_inv = view_inv
        #self.rotate_axis = view_inv[2].xyz
        #self.rotate_up = view_inv[1].xyz

        #op = self.my_target_operator(context)

        #co = Vector(op.plane_co)
        #no = Vector(op.plane_no).normalized()

        # Move
        #no_z = no
        #no_y = no_z.orthogonal()
        #no_x = no_z.cross(no_y)

        #matrix = self.gizmo_move.matrix_basis
        #matrix.identity()
        #matrix.col[0].xyz = no_x
        #matrix.col[1].xyz = no_y
        #matrix.col[2].xyz = no_z
        # The location callback handles the location.
        # `matrix.col[3].xyz = co`.

        
    
    
    
    
# Register and add to the "view" menu (required to also use F3 search "Modal Draw Operator" for quick access).
def register():
    bpy.utils.register_class(RevolveOperator)
    bpy.utils.register_class(GP_GGT_revolve_gizmo)

def unregister():
    bpy.utils.unregister_class(RevolveOperator)
    bpy.utils.unregister_class(GP_GGT_revolve_gizmo)