import bpy
import bmesh
from bpy_extras import view3d_utils
from mathutils import Vector, geometry
from mathutils.bvhtree import BVHTree
from mathutils.kdtree import KDTree




def region_2d_to_vector_3d( coord ):
    region = bpy.context.region
    rv3d = bpy.context.region_data

    from mathutils import Vector

    viewinv = rv3d.view_matrix.inverted()
    if rv3d.is_perspective:
        persinv = rv3d.perspective_matrix.inverted()
        width = region.width
        height = region.height

        out = Vector(((2.0 * coord[0] / width) - 1.0,
                      (2.0 * coord[1] / height) - 1.0,
                      -0.5
                      ))

        w = out.dot(persinv[3].xyz) + persinv[3][3]

        view_vector = ((persinv @ out) / w) - viewinv.translation
    else:
        view_vector = -viewinv.col[2].xyz

    view_vector.normalize()

    return view_vector


def region_2d_to_origin_3d( coord, clamp=None ):
    region = bpy.context.region
    rv3d = bpy.context.region_data

    viewinv = rv3d.view_matrix.inverted()

    if rv3d.is_perspective:
        origin_start = viewinv.translation.copy()
    else:
        persmat = rv3d.perspective_matrix.copy()
        dx = (2.0 * coord[0] / region.width) - 1.0
        dy = (2.0 * coord[1] / region.height) - 1.0
        persinv = persmat.inverted()
        origin_start = ((persinv.col[0].xyz * dx) +
                        (persinv.col[1].xyz * dy) +
                        persinv.translation)

        if clamp != 0.0:
            if rv3d.view_perspective != 'CAMERA':
                # this value is scaled to the far clip already
                origin_offset = persinv.col[2].xyz
                if clamp is not None:
                    if clamp < 0.0:
                        origin_offset.negate()
                        clamp = -clamp
                    if origin_offset.length > clamp:
                        origin_offset.length = clamp

                origin_start -= origin_offset

    return origin_start


def region_2d_to_location_3d( coord, depth_location ):
    rv3d = bpy.context.region_data

    from mathutils import Vector

    coord_vec = region_2d_to_vector_3d( coord )
    depth_location = Vector(depth_location)

    origin_start = region_2d_to_origin_3d( coord )
    origin_end = origin_start + coord_vec

    if rv3d.is_perspective:
        from mathutils.geometry import intersect_line_plane
        viewinv = rv3d.view_matrix.inverted()
        view_vec = viewinv.col[2].copy()
        return intersect_line_plane(origin_start,
                                    origin_end,
                                    depth_location,
                                    view_vec, 1,
                                    )
    else:
        from mathutils.geometry import intersect_point_line
        return intersect_point_line(depth_location,
                                    origin_start,
                                    origin_end,
                                    )[0]
    

def location_3d_to_region_2d( coord ) :
    region = bpy.context.region
    rv3d = bpy.context.region_data

    perspective_matrix = rv3d.perspective_matrix

    prj = perspective_matrix @ Vector((coord[0], coord[1], coord[2], 1.0))
    if prj.w > 0.0:
        width_half = region.width / 2.0
        height_half = region.height / 2.0 

        return Vector((width_half + width_half * (prj.x / prj.w),
                       height_half + height_half * (prj.y / prj.w),
                       ))
    else:
        return None


def get_point_square( coord, coords ):
        x = coord[0] - coords[0]
        y = coord[1] - coords[1]
        return x*x + y*y


def get_closest_point( coord, coords ):
    vectorAP = (coord[0] - coords[0][0], coord[1] - coords[0][1])
    vectorEdge = (coords[1][0] - coords[0][0], coords[1][1] - coords[0][1])
    sEdge = vectorEdge[0]**2 + vectorEdge[1]**2
    product = vectorEdge[0]*vectorAP[0] + vectorEdge[1]*vectorAP[1] 

    if sEdge == 0.0:
        return None
    else:
        distance = product / sEdge       

    if distance < 0.0:
        return coords[0]
    elif distance > 1.0:
        return coords[1]
    else:
        return coords[0][0] + vectorEdge[0]*distance, coords[0][1] + vectorEdge[1]*distance


class Ray:
    def __init__( self , origin : Vector , vector  : Vector ) :
        self.origin = origin.copy()
        self.vector = vector
        self.vector.normalize()


    @staticmethod
    def from_screen( context , coord : Vector) :
        #rv3d = context.region_data   
        #region = context.region
        origin = region_2d_to_origin_3d( coord )
        vector = region_2d_to_vector_3d( coord )
        return Ray(origin,vector)


    @staticmethod
    def from_world_to_screen( context , world_pos : Vector ) :
        coord = location_3d_to_region_2d( world_pos )
        if coord == None :
            return None
        return Ray.from_screen(context , coord)


    @staticmethod
    def from_line( p0 : Vector , p1 : Vector) :
        return Ray( p0 , p1 - p0 )


    def world_to_object( self , obj ) :
        matrix_inv = obj.matrix_world.inverted()
        target = self.origin + self.vector
        ray_origin_obj = matrix_inv @ self.origin
        ray_target_obj = matrix_inv @ target
        ray_direction_obj = ray_target_obj - ray_origin_obj

        return Ray( ray_origin_obj , ray_direction_obj )


    def object_to_world( self , obj ) :
        matrix = obj.matrix_world
        target = self.origin + self.vector
        ray_origin_obj = matrix @ self.origin
        ray_target_obj = matrix @ target
        ray_direction_obj = ray_target_obj - ray_origin_obj

        return Ray( ray_origin_obj , ray_direction_obj )

    
    def distance( self , ray ) :
        if self.vector.cross(ray.vector).length_squared < 0.000001 :
            return None, None, ray.origin.cross(self.origin).length

        Dv = self.vector.dot(ray.vector)
        D1 = (ray.origin-self.origin).dot(self.vector)
        D2 = (ray.origin-self.origin).dot(ray.vector)

        t1 = ( D1 - D2 * Dv ) / ( 1.0 - Dv * Dv )
        t2 = ( D2 - D1 * Dv ) / ( Dv * Dv - 1.0 )

        Q1 = self.origin + self.vector * t1
        Q2 = ray.origin + ray.vector * t2

        return Q1,Q2,(Q2 - Q1).length


    def hit_to_line( self , v0 , v1 ) :
        h0 , h1 , d = self.distance( Ray( v0 , (v1-v0) ) )

        dt =  (v0-v1).length
        d0 = (v0-h1).length
        d1 = (v1-h1).length
        if d0 > d1 and d0 >= dt :
            return 1.0
        elif d1 >= dt :
            return 0.0

        return max( 0 , min( 1 , d0 / dt ))        


    def hit_to_line_pos( self , v0 , v1 ) :
        h0 , h1 , d = self.distance( Ray( v0 , (v1-v0) ) )
        if h0 == None :
            return None
        dt =  (v0-v1).length
        d0 = (v0-h1).length
        d1 = (v1-h1).length
        if d0 > d1 and d0 >= dt :
            val =  1.0
        elif d1 >= dt :
            val =  0.0
        else :
            val = d0 / dt
        
        return v0 + (v1-v0) * val


    @property
    def invert( self ) :
        return Ray( self.origin , -self.vector )


    @property
    def x_zero( self ) :
        return Ray( self.origin , Vector( (0 , self.vector.y , self.vector.z )) )



def raycast(context, coord, measure=True): # FIXME передает индекс несуществующего элемента при использовании модификаторов
    scene = context.scene
    dg = context.evaluated_depsgraph_get()

    
    ray = Ray.from_screen(context , coord)
    

    

    success, location, normal, face_index, obj, matrix = scene.ray_cast(dg, ray.origin, ray.vector)

    dg.update()

    if success:
        return success, location, normal, face_index, obj, matrix
        
    elif measure:
        settings = context.scene.GP_scene_set
        tr_p1 = Vector(settings.tr_p1)
        ray_target = ray.origin + ray.vector
        location = geometry.intersect_line_plane(ray.origin, ray_target, tr_p1, ray.vector)

        return success, location, None, None, None, None

    else:
        ray_target = ray.origin + ray.vector
        pos_mouse = region_2d_to_location_3d(coord, ray_target)
        location = geometry.intersect_line_plane(ray.origin, ray_target, pos_mouse, ray.vector)

        return success, location, None, None, None, None



def ray_cast(context, coord):
    scene = context.scene
    region = context.region
    rv3d = context.region_data
    dg = context.evaluated_depsgraph_get()
    view_vector = view3d_utils.region_2d_to_vector_3d( region, rv3d, coord ) 
    ray_origin = view3d_utils.region_2d_to_origin_3d( region, rv3d, coord )
    return scene.ray_cast(dg, ray_origin, view_vector) # success, location, normal, face_index, obj, matrix


def get_element(obj, face, mp): # --- TODO

    bm = bmesh.from_edit_mesh(obj.data)
    bm.faces.ensure_lookup_table()

    face = bm.faces[face]
    verts_co = [obj.matrix_world @ v.co for v in face.verts]



    closest_vertex = None
    closest_distance = 10
    for vert_loc in verts_co:
        distance = (mp - vert_loc).length
        if distance < closest_distance:
            closest_distance = distance
            closest_vertex = [ vert_loc ]


    return closest_vertex, verts_co







def find_neares_from_object(context, origin, n=2, radius=1):
    obj = context.active_object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    size = len(bm.verts)
    kd = KDTree(size)

    for i, v in enumerate(bm.verts):
        kd.insert(obj.matrix_world @ v.co, i)

    kd.balance()

    fnd = kd.find(origin)
    fnd_n = kd.find_n(origin, n)
    fnd_r = kd.find_range(origin, radius)

    return fnd, fnd_n, fnd_r





















def findnearest_from_raycast(context, mp, use_boundary_vertices=False, use_boundary_edges=False): # EDBM_unified_findnearest_from_raycast
    success, co, normal, face_index, obj, matrix = raycast(context, mp)
    

    """ objs = context.objects_in_mode_unique_data
    for ob in objs:
        bm = bmesh.from_edit_mesh(ob.data)
        size = len(bm.verts)
        kd = KDTree(size)

        for i, v in enumerate(bm.verts):
            kd.insert(ob.matrix_world @ v.co, i)

        kd.balance()

        c, i, d = kd.find(co) """





    obj_i = -1
    v_i = -1
    e_i = -1
    f_i = -1
    if success:
        obj_i = obj
        v_i = -1
        e_i = -1
        f_i = face_index
    else:
        obj_i = -1
        v_i = -1
        e_i = -1
        f_i = -1
    print(obj_i, v_i, e_i, f_i)